"""
Author : Hasanain Mehmood
Contact : hasanain@aicailber.com 
"""

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns

from .common import get_cat_cols, get_num_cols

plt.style.use('fivethirtyeight')


class EDA:
    """This class have all the methods to perform complete Exploratory Data Analysis(EDA) on
       any Pandas dataframe passed as an argument on class initialization.

    Attributes:
            df (pd.DataFrame): Pandas dataframe to perform the EDA on.
    """

    def __init__(self, df: pd.DataFrame):
        """Constructs all the necessary attributes for the instance of EDA class.

        Args:
            df (pd.DataFrame): Pandas dataframe to perform the EDA on.
        """
        self.df = df

    def pie_bar(self, hue: str, kind="barh", color='red', width=18, height=6) -> object:
        """This method will plot side-by-side Pie and Bar chart with the following arguments provided:

        Args:
            hue (str): Name of the column to plot.
            kind (str, optional): Barchart kind. Defaults to "barh".
            color (str, optional): Font color. Defaults to 'red'.
            width (int, optional): Width of the plot. Defaults to 18.
            height (int, optional): Height of the plot. Defaults to 6.

        Returns:
            object: An Object which can be used to save or plot charts in any python application.
        """
        plt.figure(figsize=(width, height))
        fig, (ax1, ax2) = plt.subplots(nrows=1, ncols=2,
                                       sharey=False, figsize=(width, height))
        labels_list = []
        explode_list = []
        for i in range(self.df[hue].nunique()):
            labels_list.append(self.df[hue].value_counts().index[i])
            explode_list.append(0.01)
        ax1 = self.df[hue].value_counts().plot.pie(
            autopct="%1.0f%%", labels=labels_list, startangle=60,
            ax=ax1, wedgeprops={"linewidth": 1, "edgecolor": "k"}, explode=explode_list, shadow=True)
        ax1.set(title=f'Percentage in {hue}')

        ax2 = self.df[hue].value_counts().plot(kind=kind, ax=ax2)
        for i, j in enumerate(self.df[hue].value_counts().values):
            ax2.text(0.5, i, j, fontsize=20, color=color)
        ax2.set(title=f'Count in {hue}')
        return plt

    def histogram(self, col_to_plot="default", hue=None, kde=True, width=25, height=15) -> object:
        """This method will plot the grid or single histogram with the following arguments provided:

        Args:
            col_to_plot (list, optional): A list of the names of numerical columns. Defaults to "default".
            hue (str, optional): A categorical column name to present hue in the data. Defaults to None.
            kde (bool, optional): kernel density estimate (KDE). Defaults to True.
            width (int, optional): Width of the plot. Defaults to 25.
            height (int, optional): Height of the plot. Defaults to 15.

        Returns:
            object: An Object which can be used to save or plot charts in any python application.
        """
        sns.set(font_scale=1)
        num_cols, columns = get_num_cols(self.df, col_to_plot)
        plt.figure(figsize=(width, height))
        for i in range((len(num_cols))):
            ax = plt.subplot(int(len(num_cols)/columns), columns, i+1)
            sns.histplot(
                data=self.df, x=num_cols[i], hue=hue, element="step", kde=kde)
        plt.tight_layout()
        return plt

    def boxplot(self, col_to_plot="default", hue=None, width=20, height=10) -> object:
        """This method will plot the grid or single box plot with the following parameters provided:

        Args:
            col_to_plot (list, optional): A list of the names of numerical columns. Defaults to "default".
            hue (str, optional): A categorical column name to present hue in the data. Defaults to None.
            width (int, optional): Width of the plot. Defaults to 20.
            height (int, optional): Height of the plot. Defaults to 10.

        Returns:
            object: An Object which can be used to save or plot charts in any python application.
        """
        sns.set(font_scale=1)
        num_cols, columns = get_num_cols(self.df, col_to_plot)
        plt.figure(figsize=(width, height))
        for i in range((len(num_cols))):
            ax = plt.subplot(int(len(num_cols)/columns), columns, i+1)
            sns.boxplot(data=self.df, x=self.df[num_cols[i]], y=hue,
                        notch=True, showcaps=True,
                        flierprops={"marker": "x"},
                        boxprops={"facecolor": (.4, .6, .8, .5)},
                        medianprops={"color": "coral"},)
        plt.tight_layout()
        return plt

    def violinplot(self, hue: str, col_to_plot="default", inner="quart", width=15,
                   height=10, x_label_rotation=90, title="Violin Plot") -> object:
        """This method will plot the violinplot witht the following parameters provided:

        Args:
            hue (str): A categorical column name to present hue in the data.
            col_to_plot (str, optional): A list of the names of numerical columns. Defaults to "default".
            inner (str, optional): Representation of the datapoints in the violin interior, values could be: ['box','point','quart','stick']. Defaults to "quart".
            width (int, optional): Width of the plot. Defaults to 15.
            height (int, optional): Height of the plot. Defaults to 10.
            x_label_rotation (int, optional): Rotation of the label. Defaults to 90.
            title (str, optional): Title of the plot. Defaults to "Violin Plot".

        Returns:
            object: An Object which can be used to save or plot charts in any python application.
        """
        sns.set(font_scale=1)
        num_cols, columns = get_num_cols(self.df, col_to_plot)
        X = self.df[num_cols].drop(hue, axis=1, errors='ignore')
        y = self.df[hue]
        if hue in num_cols:
            columns_len = len(num_cols)-1
        else:
            columns_len = len(num_cols)
        data_n_2 = (X - X.mean()) / (X.std())
        data = pd.concat([y, data_n_2.iloc[:, 0:columns_len]], axis=1)
        data = data.fillna(method='ffill')
        data = pd.melt(data, id_vars=y.name,
                       var_name="features",
                       value_name='value')
        plt.figure(figsize=(width, height))
        sns.violinplot(x="features", y="value", hue=y.name,
                       data=data, split=False, inner=inner)
        plt.title(title)
        plt.xticks(rotation=x_label_rotation)
        return plt

    def jointplot(self, x: str, y: str, kind="reg", color="#ce1414", width=18, height=20) -> object:
        """This method will plot the jointplot with the following arguments provided: 

        Args:
            x (str): Numeraical column name to be plotted on x-axis
            y (str): Numeraical column name to be plotted on y-axis
            kind (str, optional): Kind of chart to plot, available selections: ['scatter','kde','hist','hex','reg', 'resid']. Defaults to "reg".
            color (str, optional): Color of the plot. Defaults to "#ce1414".
            width (int, optional): Width of the plot. Defaults to 18.
            height (int, optional): Height of the plot. Defaults to 20.

        Returns:
            object: An Object which can be used to save or plot charts in any python application.
        """
        sns.set(font_scale=1)
        plt.figure(figsize=(width, height))
        sns.jointplot(data=self.df, x=x, y=y, kind=kind, color=color)
        return plt

    def corr_heatmap(self, col_to_plot='default', width=25, height=20, title="Correlation Plot") -> object:
        """This method will plot the correlation plot with the following arguments provided:

        Args:
            col_to_plot (list, optional): A list of names of numerical columns. Defaults to 'default'.
            width (int, optional): Width of the plot. Defaults to 25.
            height (int, optional): Height of the plot. Defaults to 20.
            title (str, optional): Title of the plot. Defaults to "Correlation Plot".

        Returns:
            object: An Object which can be used to save or plot charts in any python application.
        """
        sns.set(font_scale=1)
        num_cols, columns = get_num_cols(self.df, col_to_plot)
        corr = self.df[num_cols].corr()
        mask = np.zeros_like(corr)
        mask[np.triu_indices_from(mask)] = True
        with sns.axes_style("white"):
            f, ax = plt.subplots(figsize=(width, height))
            ax = sns.heatmap(corr, mask=mask, vmax=1.0,
                             linewidths=.5, fmt='.1f', annot=True)
        plt.title(title)
        return plt

    def countplot(self, col_to_plot="default", hue=None, width=15, height=8) -> object:
        """This method will plot the count plot with the following arguments provided:

        Args:
            col_to_plot (list, optional): A list of the names of numerical columns. Defaults to "default".
            hue (_type_, optional): A categorical column name to present hue in the data. Defaults to None.
            width (int, optional): Width of the plot. Defaults to 15.
            height (int, optional): Height of the plot. Defaults to 8.

        Returns:
            object: An Object which can be used to save or plot charts in any python application.
        """
        sns.set(font_scale=1)
        cat_cols, columns = get_cat_cols(self.df, col_to_plot)
        plt.figure(figsize=(width, height))
        for i in range((len(cat_cols))):
            ax = plt.subplot(int(len(cat_cols)/columns), columns, i+1)
            sns.countplot(data=self.df, x=cat_cols[i], hue=hue)
        plt.tight_layout()
        return plt

    def pairplot(self, hue: str, width=25, height=15) -> object:
        """This method will plot the pair plot with the following arguments provided:

        Args:
            hue (str): A categorical column name to present hue in the data.
            width (int, optional): Width of the plot. Defaults to 25.
            height (int, optional): Height of the plot. Defaults to 15.

        Returns:
            object: An Object which can be used to save or plot charts in any python application.
        """
        sns.set(font_scale=1)
        plt.figure(figsize=(width, height))
        sns.pairplot(data=self.df, hue=hue)
        return plt

    def barplot(self, y: str, hue=None, col_to_plot='default', estimator='mean', width=15, height=8) -> object:
        """This method will plot the bar chart with the following arguments provided:

        Args:
            y (str): Name of the column containing numerical data to plot
            hue (str, optional): A categorical column name to present hue in the data. Defaults to None.
            col_to_plot (list, optional): A list of column names to plot. Defaults to 'default'.
            estimator (str, optional): An aggregation function, available selections: [sum, mean, meadian, std, var]. Defaults to 'mean'.
            width (int, optional): Width of the plot. Defaults to 15.
            height (int, optional): Height of the plot. Defaults to 8.

        Returns:
            object: An Object which can be used to save or plot charts in any python application.
        """
        cat_cols, columns = get_cat_cols(self.df, col_to_plot)
        plt.figure(figsize=(width, height))
        for i in range((len(cat_cols))):
            ax = plt.subplot(int(len(cat_cols)/columns), columns, i+1)
            sns.barplot(
                data=self.df, x=cat_cols[i], y=y, estimator=estimator, hue=hue)
        plt.tight_layout()
        return plt



"""
Author : Hasanain Mehmood
Contact : hasanain@aicailber.com 
"""

import pandas as pd
import plotly.express as px
import plotly.figure_factory as ff
import plotly.graph_objects as go
from plotly.subplots import make_subplots


class EDA:
    """This class have all the methods to perform complete Exploratory Data Analysis(EDA) on
       any Pandas dataframe passed as an argument on class initialization.

    Attributes:
            df (pd.DataFrame): Pandas dataframe to perform the EDA on.
    """

    def __init__(self, df:pd.DataFrame): 
        """Constructs all the necessary attributes for the instance of EDA class.

        Args:
            df (pd.DataFrame): Pandas dataframe to perform the EDA on.
        """        
        self.df = df

    def barchart(self, values: str, categories: str, aggfunc='sum', title='Bar Chart',
                 orientation='h', texposition="auto", text_color='white', width=600, height=500):
        """This method will plot a Bar chart with the following arguments provided:

        Args:
            values (str): Name of the column having numerical data to plot against the categories
            categories (str): Name of the column having the list of categories as elements
            title (str): The title of the chart. Defaults to 'Bar Chart'.
            orientation (str): To plot the chart horizontly or vertically. Defaults to 'h'.
            texposition (str): Position of the text on chart. Defaults to "auto".
            text_color (str): Text color. Defaults to 'white'.
            width (int, optional): The width of the visualization. Defaults to 600.
            height (int, optional): The height of the visualization. Defaults to 500.

        Returns:
            fig(object): An Object which can be used to save or plot charts in any python application
        """
        df = self.df.groupby([categories], as_index=False).agg(
            value=pd.NamedAgg(values, aggfunc=aggfunc)
        )
        df.columns = [categories, values]
        if orientation == 'h':
            x = df[values]
            y = df[categories]
        else:
            y = df[values]
            x = df[categories]
        fig = go.Figure(go.Bar(
            x=x,
            y=y,
            orientation=orientation,
            text=[round(val, 0) for val in df[values]], textposition=texposition, textfont=dict(color=text_color),))

        fig.update_layout(barmode='stack', xaxis={
            'categoryorder': 'total descending'})
        fig.update_layout(autosize=False, width=width,
                          height=height, title=title)
        return fig

    def heatmap(self, index: str, columns: str, values: str, aggfunc='mean', x_label=None, y_label=None, color_label=None, title='Heatmap'):
        """Thi method will plot a heatmap with the following arguments provided:

        Args:
            index (str): Name of the column to be set as index
            columns (str): Name of the column to be set as column
            values (str): Column name having numerical values
            aggfunc (str): Type of aggregation to be applied
            x_label (str): Label for x-axis. Defaults to None.
            y_label (str): Label for y-axis. Defaults to None.
            color_label (str): Color of the label. Defaults to None.
            title (str): Title of the visualization. Defaults to Heatmap.

        Returns:
            fig(object): An Object which can be used to save or plot charts in any python application
        """
        df_pivot = self.df[[index, columns, values]].pivot_table(
            index=index, columns=columns, aggfunc=aggfunc, fill_value=0)
        fig = px.imshow(df_pivot.values,
                        labels=dict(x=x_label, y=y_label, color=color_label),
                        x=[x[1] for x in df_pivot.columns],
                        y=df_pivot.index.tolist(), text_auto=True
                        )
        fig.update_traces(dict(showscale=False,
                               coloraxis=None,
                               colorscale='bluered'), selector={'type': 'heatmap'})
        fig.update_xaxes(side="top", title=title)
        return fig

    def stack_or_group_chart(self, categories: str, values: list, barmode="stack", orientation='v',
                             texposition='inside', aggfunc="mean", text_color='white', title=None):
        """This method will plot a stack or group chart using the following arguments provided:

        Args:
            categories (str): Categorical Column
            values (list): list of numerical columns to plot
            barmode (str, optional): The position of the bars for the relavent data. Defaults to "stack".
            orientation (str, optional): Orientation of the graph could be vertical or horizontal. Defaults to 'v'.
            texposition (str, optional): Position of the text labels on the plot. Defaults to 'inside'.
            aggfunc (str, optional): Aggregation function. Defaults to "mean".
            text_color (str, optional): Color of text. Defaults to 'white'.
            title (str, optional): Title of the plot. Defaults to None.

        Returns:
            fig(object): An Object which can be used to save or plot charts in any python application
        """
        df = self.df[[categories]+values].groupby(
            [categories], as_index=False).agg(aggfunc, numeric_only=True)
        data = []
        if orientation == 'v':
            for i in range(len(values)):
                data.append(go.Bar(name=values[i], x=df[categories], y=df[values[i]], orientation=orientation,
                                   text=[int(val) for val in df[values[i]]], textposition=texposition, textfont=dict(color=text_color),))
        else:
            for i in range(len(values)):
                data.append(go.Bar(name=values[i], y=df[categories], x=df[values[i]], orientation=orientation,
                                   text=[int(val) for val in df[values[i]]], textposition=texposition, textfont=dict(color=text_color),))
        fig = go.Figure(data=data)
        fig.update_layout(barmode=barmode)
        fig.update_layout(autosize=False, width=600, height=500, title=title)
        return fig

    def piechart(self, categories: str, values: str, width=600, height=500, title='Pie Chart'):
        """This method will plot a pie chart using with the following arguments provided:

        Args:
            categories (str): Categorical column name
            values (str): Numerical Column name
            width (int, optional): width of the plot. Defaults to 600.
            height (int, optional): Height of the plot. Defaults to 500.
            title (str, optional): Title of the plot. Defaults to 'Pie Chart'.

        Returns:
            fig(object): An Object which can be used to save or plot charts in any python application
        """
        fig = px.pie(data_frame=self.df, values=values, names=categories,
                     color=categories,
                     title=title,
                     template='presentation',
                     width=width,
                     height=height,
                     hole=0.5,
                     )
        return fig

    def area_chart(self, categories: str, values: str, aggfunc='mean', unit=None, width=600, height=450,  title="Area Chart"):
        """This method will plot the area chart with the following arguments provided:

        Args:
            categories (str): categorical column name.
            values (str): column name for values to plot.
            aggfunc (str, optional): Aggregation function. Defaults to 'mean'.
            unit (str, optional): Unit to be displayed with the datalabels. Defaults to None.
            width (int, optional): Width of the plot. Defaults to 600.
            height (int, optional): Height of the plot. Defaults to 450.
            title (str, optional): Title of the plot. Defaults to "Area Chart".

        Returns:
            fig(object): An Object which can be used to save or plot charts in any python application.
        """
        df = self.df.copy()
        if aggfunc == 'mean':
            df = df[[categories]+[values]
                    ].groupby([categories], as_index=False).mean()
        elif aggfunc == 'sum':
            df = df[[categories]+[values]
                    ].groupby([categories], as_index=False).sum()
        elif aggfunc == 'median':
            df = df[[categories]+[values]
                    ].groupby([categories], as_index=False).median()
        if unit == None:
            unit = " "
        else:
            unit = " "+unit
        fig = px.area(
            df,
            x=df[categories],
            y=df[values],
            text=[str(int(val))+unit for val in df[values]], title=title
        )
        fig.update_traces(mode='markers+lines+text',
                          textfont_size=12, textposition="top left")
        fig.update_layout(autosize=False, width=width, height=height,)
        return fig

    def bar_line(self, categories: str, values: list, aggfunc=['mean', 'sum'], legends=None,  width=600, height=450,
                 title="Bar Line Chart", texposition="auto", text_color="white", round_decimal=0):
        """This method will plot a bar line chart with the following arguments provided:

        Args:
            categories (str): Name of the categorical column.
            values (list): A list of the names of numerical columns.
            aggfunc (list, optional): . Defaults to ['mean','sum'].
            legends (list, optional): List containing the custom legends as elements. Defaults to None.
            width (int, optional): Width of the plot. Defaults to 600.
            height (int, optional): Height of the plot. Defaults to 450.
            title (str, optional): Title of the plot. Defaults to "Bar Line Chart".
            texposition (str, optional): Position of the data labels on the plot. Defaults to "auto".
            text_color (str, optional): Color of the text. Defaults to "white".
            round_decimal (int, optional): Decimal points to round values on. Defaults to 0.

        Returns:
            fig(object): An Object which can be used to save or plot charts in any python application.
        """
        df = self.df[[categories]+values].groupby([categories], as_index=False).agg(
            val1=pd.NamedAgg(values[0], aggfunc=aggfunc[0]),
            val2=pd.NamedAgg(values[1], aggfunc=aggfunc[1])
        )
        df.columns = [categories]+values
        if round_decimal > 0:
            line_data_labels = [round(val, round_decimal)
                                for val in df[values[0]]]
            bar_data_labels = [round(val, round_decimal)
                               for val in df[values[1]]]
        elif round_decimal == 0:
            line_data_labels = [int(val) for val in df[values[0]]]
            bar_data_labels = [int(val) for val in df[values[1]]]
        if legends is None:
            legends = values
        fig = go.Figure()
        fig.add_trace(
            go.Scatter(name=legends[0], x=df[categories], y=df[values[0]],
                       text=line_data_labels, textposition='top right',
                       mode='lines+markers+text', marker=dict(color='#5D69B1', size=7)
                       ))
        fig.add_trace(
            go.Bar(name=legends[1], x=df[categories], y=df[values[1]],
                   text=bar_data_labels, textposition=texposition, textfont=dict(color=text_color),)
        )
        # yanchor="top",y=0.99 ,x=0.01
        fig.update_layout(autosize=False, width=width,
                          height=height, title=title)
        return fig

    def histogram(self, values: str, color=None, marginal=None, hover_data=None, title='Histogram', width=800, height=450):
        """This method with plot a histogram with following parameters provided:

        Args:
            values (str): Name of the numerical column to plot.
            color (str, optional): Name of the column to differentiate data points on color. Defaults to None.
            marginal (str, optional): Type of visual to plot as marginal. Defaults to None.
            hover_data (list, optional): A list of column names to display the data on hover to the marginal visual. Defaults to None.
            title (str, optional): Title of the plot. Defaults to 'Histogram'.
            width (int, optional): Width of the plot. Defaults to 800.
            height (int, optional): Height of the plot. Defaults to 450.

        Returns:
            fig(object): An Object which can be used to save or plot charts in any python application.
        """
        fig = px.histogram(self.df, x=values, y=values, color=color,
                           marginal=marginal, histfunc='count', hover_data=hover_data)
        fig.update_layout(autosize=False, width=width,
                          height=height, title=title)
        return fig

    def distplot(self, values: list, width=800, height=450, title='Distribution Plot'):
        """Thi method will plot the distribution plot with the following parameters provided:

        Args:
            values (list): A list of the names of the numerical columns to plot.
            width (int, optional): Width of the plot. Defaults to 800.
            height (int, optional): Height of the plot. Defaults to 450.
            title (str, optional): Title of the plot. Defaults to 'Distribution Plot'.

        Returns:
            fig(object): An Object which can be used to save or plot charts in any python application.
        """
        df = self.df[values]
        fig = ff.create_distplot(
            [df[c] for c in df.columns], df.columns, bin_size=.25)
        fig.update_layout(autosize=False, width=width,
                          height=height, title=title)
        return fig

    def combined_corr(self, x_values: str, y_values: str, color=None, hover_name=None, size=None, title="Combined Correlation Plot"):
        """This method will plot the combined correlation plot using the following arguments:

        Args:
            x_values (str): Name of the column to plot against x_axis.
            y_values (str): Name of the column to plot against y_axis.
            color (str, optional): Name of the column to differentiate data points on color. Defaults to None.
            hover_name (str, optional): Column name to display as hover name. Defaults to None.
            size (str, optional): Name of the column to be presented as the size of the data point. Defaults to None.
            title (str, optional): Title of the plot. Defaults to "Combined Correlation Plot".

        Returns:
            fig(object): An Object which can be used to save or plot charts in any python application.
        """
        fig = px.scatter(self.df, x=x_values, y=y_values, color=color,
                         marginal_x="box", marginal_y='violin', hover_name=hover_name,
                         size=size, title=title)
        return fig

    def multivar_bubble_chart(self, categories: list, values: list, aggfunc="mean", hover_name=None, width=600, height=450, title="Bubble Chart"):
        """This method will plot a multivariate bubble chart with the following arguments provided:

        Args:
            categories (list): List of the names of categorical columns to plot.
            values (list): List of the names of values columns to plot.
            aggfunc (str, optional): Aggregation function. Defaults to "mean".
            hover_name (_type_, optional): Name of the category to be displayed on hover. Defaults to None.
            width (int, optional): With of the plot. Defaults to 600.
            height (int, optional): Height of the plot. Defaults to 450.
            title (str, optional): Title of the plot. Defaults to "Bubble Chart".

        Returns:
            fig(object): An Object which can be used to save or plot charts in any python application.
        """
        df = self.df.groupby(categories, as_index=False).agg(
            aggfunc, numeric_only=True)[categories+values]
        fig = px.scatter(df, x=df[categories[0]], y=df[values[0]],
                         size=df[values[1]], color=df[categories[1]],
                         hover_name=hover_name, size_max=20)
        fig.update_traces(dict(showlegend=True))
        fig.update_layout(autosize=False, width=width, height=height,)
        fig.update_xaxes(side="top", title=title)
        return fig

    def stacked_area_chart(self, time: str, values: list, aggfunc='mean',
                           legend=None, unit=None, width=600, height=500, title="Stacked Area Chart"):
        """This method will plot a stacked area chart with following arguments provided:

        Args:
            time (str): A time or any relevant column name to be presented on x-axis.
            values (list): A list of numerical column to be ploted against the x-axis.
            aggfunc (str, optional): Aggregation function. Defaults to 'mean'.
            legend (_type_, optional): List of legend names. Defaults to None.
            unit (_type_, optional): Name of the measurement unit. Defaults to None.
            width (int, optional): Width of the column. Defaults to 600.
            height (int, optional): Height of the column. Defaults to 500.
            title (str, optional): Title of the plot. Defaults to "Stacked Area Chart".

        Returns:
            fig(object): An Object which can be used to save or plot charts in any python application.
        """
        if unit == None:
            unit = " "
        else:
            unit = " "+unit

        df = self.df.groupby(time, as_index=False)\
            .agg(aggfunc, numeric_only=True)[[time]+values]

        fig = go.Figure()
        if legend is None:
            legend = values
        for i in range(len(values)):
            fig.add_trace(go.Scatter(
                x=df[time], y=df[values[i]],
                hoverinfo='x+y',
                mode='lines+markers+text',
                text=[str(round(i, 0))+unit for i in df[values[i]]],
                name=legend[i],
                textposition='top left',
                line=dict(width=0.5),
                marker=dict(size=7),
                stackgroup='one'  # define stack group
            ))

        fig.update_layout(autosize=False, width=width,
                          height=height, title=title)
        return fig

    def scatterplot(self, values: list, size=None, color=None, hover_name=None, width=600, height=450, title='Scatter chart'):
        """This method will plot a scatter plot with the following parameters provided:

        Args:
            values (list): List of the names of numerical columns to be plotted against each other.
            size (str, optional): Name of the column to be presented as the size of the data point. Defaults to None.
            color (str, optional): Name of the column to differentiate data points on color. Defaults to None.
            hover_name (str, optional): Name of the column to be presented as the name on hover. Defaults to None.
            width (int, optional): Width of the plot. Defaults to 600.
            height (int, optional): Height of the plot. Defaults to 450.
            title (str, optional): Title of the plot. Defaults to 'Scatter chart'.

        Returns:
            fig(object): An Object which can be used to save or plot charts in any python application.
        """
        fig = px.scatter(self.df, x=values[0], y=values[1],
                         size=size, color=color,
                         hover_name=hover_name)
        fig.update_layout(autosize=False, width=width,
                          height=height, title=title)

        return fig

    def facetgrid(self, categories: str, values: str, facet_row: str, facet_col: str, color=None,
                  aggfunc='mean', barmode='stack', width=600, height=450, title='Facetgrid'):
        """This method will plot the facet grid with the following arguments provided:

        Args:
            categories (str): Name of the categorical column.
            values (str): Name of the numerical column.
            facet_row (_type_): Name of the column to be plotted as a facet row.
            facet_col (_type_): Name of the column to be plotted as a facet column.
            color (str, optional): Name of the column to differentiate data points on color. Defaults to None.
            aggfunc (str, optional): Aggregation function. Defaults to 'mean'.
            barmode (str, optional): Chart could be plotted with the stack of group of bars. Defaults to 'stack'.
            width (int, optional): Width of the plot. Defaults to 600.
            height (int, optional): Height of the plot. Defaults to 450.
            title (str, optional): Title of the plot. Defaults to 'Facetgrid'.

        Returns:
            fig(object): An Object which can be used to save or plot charts in any python application.
        """
        cols_list = [categories]+[color]+[facet_col]+[facet_row]+[values]
        df = self.df.groupby(cols_list[:-1], as_index=False)\
            .agg(aggfunc, numeric_only=True)[cols_list]
        fig = px.bar(df, x=categories, y=values, color=color, barmode=barmode,
                     facet_row=facet_row, facet_col=facet_col,
                     category_orders={facet_row: df[facet_row].tolist(),
                                      facet_col: df[facet_col].tolist()})
        fig.update_layout(autosize=False, width=width,
                          height=height, title=title)
        return fig

    def pareto_chart(self, categories: str, values: str, unit=None,  y_label=None, width=600, height=500, title='Pareto Chart'):
        """This method will plot a pareto chart with the following arguments provided:

        Args:
            categories (str): Categorical column name.
            values (str): Numerical column name.
            unit (str, optional): Unit of measurement to plot with data labels. Defaults to None.
            y_label (str, optional): A custom name for the y-label. Defaults to None.
            width (int, optional): Width of the plot. Defaults to 600.
            height (int, optional): Height of the plot. Defaults to 500.
            title (str, optional): Title of the plot. Defaults to 'Pareto Chart'.

        Returns:
            fig(object): An Object which can be used to save or plot charts in any python application.
        """
        if y_label is None:
            y_label = values
        if unit == None:
            unit = " "
        else:
            unit = " "+unit
        df = self.df.groupby([categories], as_index=False).agg(
            "sum", numeric_only=True)[[categories]+[values]]
        df.sort_values(by=values, ascending=False, inplace=True)
        trace_0 = go.Bar(
            x=df[categories],
            y=df[values],
            marker=dict(color=df[values], coloraxis="coloraxis"),
            text=[str(int(i))+unit for i in df[values]],
            textposition='auto'
        )

        df["cumulative_%"] = 100 * (df[values].cumsum() / df[values].sum())

        trace_1 = go.Scatter(
            x=df[categories],
            y=df["cumulative_%"],
            text=[str(round(i, 0))+"%" for i in df['cumulative_%']],
            mode="lines+markers+text",
            textposition="top left"
        )

        fig = make_subplots(specs=[[{"secondary_y": True}]])

        fig.add_trace(trace_0)

        fig.add_trace(trace_1, secondary_y=True)

        fig.update_layout(
            title=title,
            autosize=False,
            width=width, height=height,
            yaxis={"title": y_label},
            showlegend=False,
            coloraxis_showscale=False
        )

        return fig



import numpy as np

def get_num_cols(df, col_to_plot):
    if col_to_plot == "default":
        num_cols = df.select_dtypes(include=[np.number]).columns.tolist()
    else:
        if not isinstance(col_to_plot, list):
            num_cols = [col_to_plot]
        else:
            num_cols = col_to_plot
    try:
        columns = [n for n in range(1, 10) if len(num_cols) % n == 0][1]  # columns for plot
    except:
        columns = 1
    return num_cols, columns


def get_cat_cols(df, col_to_plot):
    if col_to_plot == "default":
        cat_cols = df.select_dtypes(
            include=['category', 'object']).columns.tolist()  # columns in df
    else:
        if not isinstance(col_to_plot, list):
            cat_cols = [col_to_plot]
        else:
            cat_cols = col_to_plot
    try:
        columns = [n for n in range(1, 10) if len(
            cat_cols) % n == 0][1]  # columns for plot
    except:
        columns = 1
    return cat_cols, columns




import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

PROJECT_NAME = "vizpool"
USERNAME = "Hassi34"

setuptools.setup(
    name=f"{PROJECT_NAME}",
    version="0.0.8",
    author= USERNAME,
    author_email="hasanain@aicaliber.com",
    description="A Python Library with Low-Code support for Basic to Advance Static & Interactive Visualizations",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url=f"https://github.com/{USERNAME}/{PROJECT_NAME}",
    project_urls={
        "Bug Tracker": f"https://github.com/{USERNAME}/{PROJECT_NAME}/issues",
    },
    classifiers=[
            'Intended Audience :: Science/Research',
            'Programming Language :: Python :: 3.7',
            'Programming Language :: Python :: 3.8',
            'Programming Language :: Python :: 3.9',
            'Programming Language :: Python :: 3.10',
            'License :: OSI Approved :: MIT License',
            'Topic :: Scientific/Engineering :: Visualization',
            'Operating System :: OS Independent'
    ],
    package_dir={"": "src"},
    packages=setuptools.find_packages(where="src"),
    python_requires=">=3.7",
    install_requires = [
        "plotly >= 5.8.0",
        "numpy >= 1.22.4",
        "pandas >= 1.3.5",
        "seaborn >= 0.11.1",
        "scikit-learn==1.1.2",
        "kaleido==0.0.3"
    ]
)   


"""
Author : Hasanain Mehmood
Contact : hasanain@aicailber.com 
"""

import pickle
from array import array
from typing import Union

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
from sklearn.metrics import confusion_matrix, roc_auc_score, roc_curve

plt.style.use('fivethirtyeight')


class Evaluation:
    """This class will provide the common Machine learning evaluation plots.

    Attributes:
            y_val (array): Validation or test data to be predicted.
    """

    def __init__(self, y_val: Union[pd.Series, array]):
        """Constructs all the necessary attributes for the instance of Evaluation class.

        Args:
            y_val (Union[pd.Series, array]): Validation or test data to be predicted.
        """
        self.y_val = y_val

    def confusion_matrix(self, y_predicted: array, target_names=None, normalize=True, width=20, height=20, title="Confusion Matrix") -> object:
        """This method will plot the confusion matrix with the following arguments provided:

        Args:
            y_predicted (array): An array of predicted values.
            target_names (list, optional): List of target names. Defaults to None.
            normalize (bool, optional): Whether to normalize the values or not. Defaults to True.
            width (int, optional): Width of the plot. Defaults to 20.
            height (int, optional): Height of the plot. Defaults to 20.
            title (str, optional): Title of the plot. Defaults to "Confusion Matrix".

        Returns:
            object: An Object which can be used to save or plot charts in any python application.
        """
        sns.set(font_scale=1.2)
        if target_names is not None and len(target_names) > 0:
            y_unique = target_names
        else:
            y_unique = list(set(self.y_val))
        classes = len(y_unique)
        cm = confusion_matrix(self.y_val, y_predicted)
        plt.figure(figsize=(width, height))
        if normalize:
            con = np.zeros((classes, classes))
            for x in range(classes):
                for y in range(classes):
                    con[x, y] = cm[x, y]/np.sum(cm[x, :])
            sns.heatmap(con, annot=True, fmt='.2', cmap='Blues',
                        xticklabels=y_unique, yticklabels=y_unique)
        else:
            sns.heatmap(cm, annot=True, cmap='Blues',
                        xticklabels=y_unique, yticklabels=y_unique, fmt='g')
        plt.title(title)
        return plt

    def feature_importance(self, classifier: object, X_train: pd.DataFrame, y_train: Union[pd.DataFrame, pd.Series, array],
                           width=15, height=10, title='Feature Importance') -> object:
        """This method will plot the chart for the feature importances with a trianed classifier of pipeline while following arguments being provided:

        Args:
            classifier (object): A trained classifier or a pipeline.
            X_train (pd.DataFrame): A pandas dataframe.
            y_train (Union[pd.DataFrame, pd.Series, array]): Class labels of the dataframe.
            width (int, optional): Width of the plot. Defaults to 15.
            height (int, optional): Height of the plot. Defaults to 10.
            title (str, optional): Title of the plot. Defaults to 'Feature Importance'.

        Returns:
            object: An Object which can be used to save or plot charts in any python application.
        """
        sns.set(font_scale=1)
        classifier.fit(X_train, y_train)
        try:
            clf_name = classifier["classifier"].__class__.__name__
        except:
            clf_name = classifier.__class__.__name__
        try:
            weights = classifier.coef_
        except:
            try:
                weights = classifier.feature_importances_
            except:
                try:
                    weights = classifier["classifier"].coef_
                except:
                    try:
                        weights = classifier["classifier"].feature_importances_
                    except:
                        print(
                            f"\n!!! No Coefficients to plot for {clf_name}, skipping feature importance plot for {clf_name}")
                        return None
        weights = np.array(weights).flatten()
        try:
            weights_df = pd.DataFrame({
                'columns': X_train.columns,
                'weight': weights
            }).sort_values('weight', ascending=False)
        except:
            weights_df = pd.DataFrame({
                'columns': [i for i in range(len(weights))],
                'weight': weights
            }).sort_values('weight', ascending=False)
        weights_df["abs_value"] = weights_df["weight"].apply(lambda x: abs(x))
        weights_df["colors"] = weights_df["weight"].apply(
            lambda x: "green" if x > 0 else "red")
        weights_df = weights_df.sort_values("abs_value", ascending=False)

        fig, ax = plt.subplots(1, 1, figsize=(width, height))
        sns.barplot(x="columns",
                    y="weight",
                    data=weights_df.head(30),
                    palette=weights_df.head(30)["colors"])
        ax.set_xticklabels(ax.get_xticklabels(), rotation=45, fontsize=8)
        ax.set_ylabel("Coef", fontsize=12)
        ax.set_xlabel("Feature Name", fontsize=10)
        plt.title(title, fontweight='bold', fontsize=15)
        return plt

    def auc_roc_plot(self, X_val: pd.DataFrame, classifiers: list, classifier_names: list, width=15, height=10, title='ROC Curve Analysis') -> object:
        """This method will plot the ROC curve analysis with the following parameters provided:

        Args:
            X_val (pd.DataFrame): Features of test or validation dataset.
            classifiers (list): List of classifiers or pipelines.
            classifier_names (list): List of classifier names.
            width (int, optional): Width of the plot. Defaults to 15.
            height (int, optional): Height of the plot. Defaults to 10.
            title (str, optional): Title of the plot. Defaults to 'ROC Curve Analysis'.

        Returns:
            object: An Object which can be used to save or plot charts in any python application.
        """
        sns.set(font_scale=1)
        if len(np.unique(self.y_val)) > 2:
            print("AUC ROC Plot supports binary classification only, it is not available for Multiclass classification")
            return None
        result_dict = {'classifiers': [], 'fpr': [], 'tpr': [], 'auc': []}

        for model, clf_name in zip(classifiers, classifier_names):
            try:
                with open(model, 'rb') as f:
                    model = pickle.load(f)
            except:
                pass
            try:
                yproba = model.predict_proba(X_val)[::, 1]
            except AttributeError:
                try:
                    yproba = model._predict_proba_lr(X_val)[::, 1]
                except AttributeError:
                    yproba = None
            try:
                fpr, tpr, _ = roc_curve(self.y_val,  yproba)
                auc = roc_auc_score(self.y_val, yproba)

                result_dict['classifiers'].append(clf_name)
                result_dict['fpr'].append(fpr)
                result_dict['tpr'].append(tpr)
                result_dict['auc'].append(auc)
            except:
                pass
        result_table = pd.DataFrame(result_dict).set_index('classifiers')
        fig = plt.figure(figsize=(width, height))

        for i in result_table.index:
            plt.plot(result_table.loc[i]['fpr'],
                     result_table.loc[i]['tpr'],
                     label=f"{i}, AUC={result_table.loc[i]['auc']:.3f}")

        plt.plot([0, 1], [0, 1], color='orange', linestyle='--')

        plt.xticks(np.arange(0.0, 1.1, step=0.1))
        plt.xlabel("Flase Positive Rate", fontsize=15)
        plt.yticks(np.arange(0.0, 1.1, step=0.1))
        plt.ylabel("True Positive Rate", fontsize=15)
        plt.title(title, fontweight='bold', fontsize=15)
        plt.legend(prop={'size': 13}, loc='lower right')
        return plt

    def residplot(self, y_predicted: Union[array, pd.Series], color="#ce1414", width=18, height=20) -> object:
        """This method will plot the residual plot with the following arguments provided: 

        Args:
            y_predicted (Union[pd.Series, array]): An array or Pandas series representing the predicted values.
            color (str, optional): Color of the plot. Defaults to "#ce1414".
            width (int, optional): Width of the plot. Defaults to 18.
            height (int, optional): Height of the plot. Defaults to 20.

        Returns:
            object: An Object which can be used to save or plot charts in any python application.
        """
        sns.set(font_scale=1)
        plt.figure(figsize=(width, height))
        sns.jointplot(x=self.y_val, y=y_predicted, kind='resid', color=color)
        return plt